(* Mathematica Package *)

(* Created by the Wolfram Workbench Dec 8, 2010 *)

BeginPackage["PubliconToEndNoteBibConverter`"]
(* Exported symbols added here with SymbolName::usage *) 

 ConvertPubliconXMLToEndNoteXML::usage="ConvertPubliconXMLToEndNoteXML[PubliconBibXMLFilePath_String]--The function takes a single argument, the file path of a Publicon XML bibliographical database. It converts the file to EndNote XML and writes the output to a file in the same directory as the source. The output file path is returned. The output file has the same name as the source file but .xml is replaced with _EN.xml.";
 PubliconToEndNoteUI::usage="PubliconToEndNoteUI[]--User interface for selecting a Publicon XML source file and running the conversion function: PubliconToEndNoteBibConverter`ConvertPubliconXMLToEndNoteXML. The output file path is returned.";
 
Begin["`Private`"]
(* Implementation of the package *) 
 
 
 PubliconToEndNoteUI[]:=
 Module[{
	endNoteXMLFl,
	enFlNm,
	enDirPath,
	progDialog,
	$done = False
},
		(* Clear variable used that stores the name of the xml source file: PblToEnConv`pblXMLFl*)
		Clear[PblToEnConv`pblXMLFl];

		progDialog = CreateDialog[{
	ProgressIndicator[Dynamic[Clock[Infinity]], Indeterminate],
	
	DefaultButton[
				   Tooltip["Close", 
						   "Closes progress dialog, but does not Cancel evaluation", 
						   TooltipDelay -> 0.38],
				   DialogReturn[]]
 },
								  WindowTitle -> "Processing\[Ellipsis]Please Wait!",
								  WindowSize -> {250, FitAll}
								  ];
		
		PblToEnConv`pblXMLFl = SystemDialogInput["FileOpen"];
		
		(*Wait for SystemDialogInput to resolve.*)
		While[Not@PblToEnConv`$done,
			  If[ValueQ[PblToEnConv`pblXMLFl],
				 $done = True
				 ]];
		
		
		
		Catch[
			  If[PblToEnConv`pblXMLFl === $Canceled,
				 NotebookClose[progDialog];
				 Throw[PblToEnConv`pblXMLFl]
				 ];
			  
			  endNoteXMLFl = 
			  PubliconToEndNoteBibConverter`ConvertPubliconXMLToEndNoteXML[
																		   PblToEnConv`pblXMLFl];
			  
			  If[endNoteXMLFl === $Failed,
				 NotebookClose[progDialog];
				 Throw[System`MessageDialog[
											"Conversion Failed!!! Unable to convert the file: " ~~ 
											PblToEnConv`pblXMLFl ~~ " to EndNote XML.", 
											WindowTitle -> "EndNote Conversion Failed!"]]
				 ];
			  
			  enFlNm = 
			  Take[StringSplit[endNoteXMLFl, $PathnameSeparator], -1] // First;
			  enDirPath = 
			  FileNameJoin[
						   Drop[StringSplit[endNoteXMLFl, $PathnameSeparator], -1]];
			  
			  If[StringMatchQ[endNoteXMLFl, $PathnameSeparator ~~ __],
				 enDirPath = $PathnameSeparator ~~ enDirPath
				 ];
			  
			  NotebookClose[progDialog];
			  
			  With[{
			flPath = endNoteXMLFl,
			flNm = enFlNm,
			dirPath = enDirPath
		},
				   CreateDialog[
								{
									Row[{TextCell["The EndNote XML file: "],
										Button[flNm,
											   DialogReturn[SystemOpen[dirPath]
															],
											   BaseStyle -> "Hyperlink",
											   Appearance -> "Frameless"
											   ],
										TextCell[
												 " was written in the directory below. Click the link to open the directory."]}],
									Button[dirPath,
										   DialogReturn[SystemOpen[dirPath]
														],
										   BaseStyle -> "Hyperlink",
										   Appearance -> "Frameless"
										   ],
									CancelButton[]
								},
								
								WindowTitle -> "EndNote XML File Was Created"
								]
				   ];
			  endNoteXMLFl
			  ]
		]
 
 
 ConvertPubliconXMLToEndNoteXML[pblXMLFl_String]:=Module[{endNoteXMLFlNm=StringDrop[pblXMLFl,-4]<>"_EN.xml", pblXML, endNoteXML, endNoteXMLStr, exprt1, exprt2},
														 Catch[
															   If[Not@StringMatchQ[pblXMLFl, "*.xml", IgnoreCase->True],
																  Message[ConvertPubliconXMLToEndNoteXML::BadInputFileType, pblXMLFl];
																  Throw[$Failed]
																  ];
															   
															   pblXML = Import[pblXMLFl];
															   
															   If[pblXML===$Failed,
																  Message[ConvertPubliconXMLToEndNoteXML::InvalidInputFile, pblXMLFl];
																  Throw[$Failed]
																  ];
															   
															   (*Check for unsupported RefTypes*)
															   CheckRefTypes[pblXML];
															   
															   (*Check for unsupported data fields*)															   
															   CheckFieldTypes[pblXML];
															   
															   endNoteXML = ConvertRefDatabasePass2[ConvertRefDatabase[pblXML]];
															   
															   exprt1=Export[endNoteXMLFlNm, endNoteXML, "XML", CharacterEncoding -> "UTF-8"];
															   
															   If[exprt1===$Failed,
																  Message[ConvertPubliconXMLToEndNoteXML::XMLExportFailed, "1", endNoteXMLFlNm];
																  Throw[$Failed]
																  ];
															   
															   (* Reimport the export xml file to remove linebreaks between inline <style> tagged text. 
																Would have used ExportString instead of Export, above, but CharacterEncoding was not working properly. *)
															   endNoteXMLStr = Import[endNoteXMLFlNm, "Text"];
															   
															   endNoteXMLStr=StringReplace[endNoteXMLStr, {a:"<"~~b:WordCharacter..~~c:">"~~WhitespaceCharacter ..~~d:"<style" :> a~~b~~c~~d, a:"</style>"~~WhitespaceCharacter..~~b:"<":>a~~b}];
															   
															   exprt2=Export[endNoteXMLFlNm, endNoteXMLStr, "Text", CharacterEncoding->"UTF-8"];
															   
															   If[exprt2===$Failed,
																  (
																   Message[ConvertPubliconXMLToEndNoteXML::XMLExportFailed, "2", endNoteXMLFlNm];
																   $Failed
																   ),
																  exprt2
																  ]
															   
															   ]
														 ]
 
 ConvertPubliconXMLToEndNoteXML::BadInputFileType="Invalid input file type: `1`; Expected file name with \".xml\" extension.";
 ConvertPubliconXMLToEndNoteXML::InvalidInputFile="The file: `1` is either invalid or does not exist.";

 
 ConvertRefDatabase[
					XMLObject["Document"][{XMLObject["Declaration"]["Version" -> _], 
	 XMLObject["Doctype"]["bibliography", "System" -> _]}, 
										  XMLElement["bibliography", {}, recs_List], {}, "Valid" -> True]] :=
 Module[
		{},
		XMLObject[
				  "Document"][{XMLObject["Declaration"]["Version" -> "1.0", 
														"Encoding" -> "UTF-8"]}, 
							  XMLElement["xml", {}, {XMLElement["records", {},
																ConvertRefRecord[#] & /@ recs
																]}], {}]
		]
 
 ConvertRefRecord[
				  XMLElement["reference", {"id" -> _, "type" -> refType_}, 
							 cont_List]] := Module[
												   {},
												   XMLElement["record", {},
															  Prepend[
																	  ConvertRefField[#] & /@ cont,
																	  MakeRefType[refType]
																	  ]
															  ]
												   ]
 
 ConvertRefField[XMLElement["item", {"type" -> fldNm_}, {cont__}]] := 
 Module[
		{
			contLst = {cont}
		},
		contLst = 
		contLst /. 
		styTxt : XMLElement[_String, {}, {_String}] :> 
		ConvertFormattedText[styTxt];
		
		XMLElement[fldNm, {}, contLst]
		]
 
 ConvertRefField[XMLElement["item", {"type" -> fldNm_}, {}]] :=
 
 XMLElement[fldNm, {}, {}]
 
 ConvertFormattedText[XMLElement[sty_, {}, cont_List]] :=
 Which[
	   sty === "it", 
	   XMLElement[
				  "style", {"face" -> "italic", "font" -> "Helvetica", 
					  "size" -> "100%"}, cont],
	   sty === "b" || sty === "bd", 
	   XMLElement[
				  "style", {"face" -> "bold", "font" -> "Helvetica", 
					  "size" -> "100%"}, cont],
	   sty === "bit" || sty === "bdit", 
	   XMLElement[
				  "style", {"face" -> "bolditalic", "font" -> "Helvetica", 
					  "size" -> "100%"}, cont],
	   True, XMLElement[
						"style", {"face" -> "normal", "font" -> "Helvetica", 
							"size" -> "100%"}, cont]
	   ]
 
 MakeRefType[refType_String] := Module[
									   {},
									   
									   Which[
											 refType === "Bill", 
											 XMLElement["ref-type", {"name" -> "Bill"}, {"4"}], 
											 refType === "Case", 
											 XMLElement["ref-type", {"name" -> "Case"}, {"7"}], 
											 refType === "Hearing", 
											 XMLElement["ref-type", {"name" -> "Hearing"}, {"14"}], 
											 refType === "Statute", 
											 XMLElement["ref-type", {"name" -> "Statute"}, {"31"}], 
											 refType === "Electronic Article", 
											 XMLElement["ref-type", {"name" -> "Electronic Article"}, {"43"}], 
											 refType === "Article", 
											 XMLElement["ref-type", {"name" -> "Journal Article"}, {"17"}], 
											 refType === "Magazine Article", 
											 XMLElement["ref-type", {"name" -> "Magazine Article"}, {"19"}], 
											 refType === "Newspaper Article", 
											 XMLElement["ref-type", {"name" -> "Newspaper Article"}, {"23"}], 
											 refType === "Artwork", 
											 XMLElement["ref-type", {"name" -> "Artwork"}, {"2"}], 
											 refType === "Blog", 
											 XMLElement["ref-type", {"name" -> "Blog"}, {"56"}], 
											 refType === "WholeBook", 
											 XMLElement["ref-type", {"name" -> "Book"}, {"6"}], 
											 refType === "Edited Book", 
											 XMLElement["ref-type", {"name" -> "Edited Book"}, {"28"}], 
											 refType === "Electronic Book", 
											 XMLElement["ref-type", {"name" -> "Electronic Book"}, {"44"}], 
											 refType === "Film or Broadcast", 
											 XMLElement["ref-type", {"name" -> "Film or Broadcast"}, {"21"}], 
											 refType === "Catalog", 
											 XMLElement["ref-type", {"name" -> "Catalog"}, {"8"}], 
											 refType === "Personal Communication", 
											 XMLElement[
														"ref-type", {"name" -> "Personal Communication"}, {"26"}], 
											 refType === "Aggregated Database", 
											 XMLElement["ref-type", {"name" -> "Aggregated Database"}, {"55"}], 
											 refType === "OnlineSource", 
											 XMLElement["ref-type", {"name" -> "Online Database"}, {"45"}], 
											 refType === "Dataset", 
											 XMLElement["ref-type", {"name" -> "Dataset"}, {"59"}], 
											 refType === "Dictionary", 
											 XMLElement["ref-type", {"name" -> "Dictionary"}, {"52"}], 
											 refType === "Government Document", 
											 XMLElement["ref-type", {"name" -> "Government Document"}, {"46"}], 
											 refType === "Encyclopedia", 
											 XMLElement["ref-type", {"name" -> "Encyclopedia"}, {"53"}], 
											 refType === "Equation", 
											 XMLElement["ref-type", {"name" -> "Equation"}, {"39"}], 
											 refType === "Figure", 
											 XMLElement["ref-type", {"name" -> "Figure"}, {"37"}], 
											 refType === "Grant", 
											 XMLElement["ref-type", {"name" -> "Grant"}, {"54"}], 
											 refType === "Manuscript", 
											 XMLElement["ref-type", {"name" -> "Manuscript"}, {"36"}], 
											 refType === "Map", 
											 XMLElement["ref-type", {"name" -> "Map"}, {"20"}], 
											 refType === "Audiovisual Material", 
											 XMLElement["ref-type", {"name" -> "Audiovisual Material"}, {"3"}], 
											 refType === "Online Multimedia", 
											 XMLElement["ref-type", {"name" -> "Online Multimedia"}, {"48"}], 
											 refType === "Web Page", 
											 XMLElement["ref-type", {"name" -> "Web Page"}, {"12"}], 
											 refType === "Pamphlet", 
											 XMLElement["ref-type", {"name" -> "Pamphlet"}, {"24"}], 
											 refType === "Conference Paper", 
											 XMLElement["ref-type", {"name" -> "Conference Paper"}, {"47"}], 
											 refType === "Patent", 
											 XMLElement["ref-type", {"name" -> "Patent"}, {"25"}], 
											 refType === "Conference Proceedings", 
											 XMLElement[
														"ref-type", {"name" -> "Conference Proceedings"}, {"10"}], 
											 refType === "Computer Program", 
											 XMLElement["ref-type", {"name" -> "Computer Program"}, {"9"}], 
											 refType === "Legal Rule or Regulation", 
											 XMLElement[
														"ref-type", {"name" -> "Legal Rule or Regulation"}, {"50"}], 
											 refType === "Report", 
											 XMLElement["ref-type", {"name" -> "Report"}, {"27"}], 
											 refType === "InBook", 
											 XMLElement["ref-type", {"name" -> "Book Section"}, {"5"}], 
											 refType === "Electronic Book Section", 
											 XMLElement[
														"ref-type", {"name" -> "Electronic Book Section"}, {"60"}], 
											 refType === "Serial", 
											 XMLElement["ref-type", {"name" -> "Serial"}, {"57"}], 
											 refType === "Standard", 
											 XMLElement["ref-type", {"name" -> "Standard"}, {"58"}], 
											 refType === "Chart or Table", 
											 XMLElement["ref-type", {"name" -> "Chart or Table"}, {"38"}], 
											 refType === "Ancient Text", 
											 XMLElement["ref-type", {"name" -> "Ancient Text"}, {"51"}], 
											 refType === "Thesis", 
											 XMLElement["ref-type", {"name" -> "Thesis"}, {"32"}], 
											 refType === "Classical Work", 
											 XMLElement["ref-type", {"name" -> "Classical Work"}, {"49"}], 
											 refType === "Preprint", 
											 XMLElement["ref-type", {"name" -> "Unpublished Work"}, {"34"}],
											 refType === "Note" || True, 
											 XMLElement["ref-type", {"name" -> "Generic"}, {"13"}]
											 ]
									   ]
 
 ConvertRefDatabasePass2[
						 XMLObject[
								   "Document"][{XMLObject["Declaration"]["Version" -> ver_, 
																		 "Encoding" -> enc_]}, 
											   XMLElement["xml", {}, {XMLElement["records", {}, recs_List]}], 
											   other___]] := 
 XMLObject[
		   "Document"][{XMLObject["Declaration"]["Version" -> ver, 
												 "Encoding" -> enc]}, 
					   XMLElement["xml", {}, {XMLElement["records", {},
														 ConvertAuthor[recs] // ConvertTitle // ConvertDate // 
														 ConvertPublisherInfo // ConvertFlatFields // 
														 ConvertPageRange // ConvertURL // FixSecondaryTitle // 
														 RemoveMathML // FixPlainText
														 ]}], other]
 
 ConvertAuthor[xpr_] :=
 (
  (
   xpr //.
   currXpr : {pre___,
	   XMLElement[
				  
				  ae1_String?(StringMatchQ[#, "Author*N"] || 
							  StringMatchQ[#, "Editor*N"] &), {}, a_List],
	   XMLElement[
				  
				  ae2_String?(StringMatchQ[#, "Author*N"] || 
							  StringMatchQ[#, "Editor*N"] &), {}, b_List],
	   XMLElement[
				  
				  ae3_String?(StringMatchQ[#, "Author*N"] || 
							  StringMatchQ[#, "Editor*N"] &), {}, c_List], post___} :>
   If[
	  StringDrop[ae1, -2] === StringDrop[ae2, -2] === 
	  StringDrop[ae3, -2],
	  {pre,
		  XMLElement["contributors", {}, {
			  XMLElement[Which[
							   StringTake[ae1, 6] === "Editor", "secondary-authors",
							   True, "authors"
							   ], {}, {
				  XMLElement["author", {"first-name" -> Which[
															  StringTake[ae1, -2] === "FN" && Length[a] == 1, 
															  First[a],
															  StringTake[ae2, -2] === "FN" && Length[b] == 1, 
															  First[b],
															  StringTake[ae3, -2] === "FN" && Length[c] == 1, 
															  First[c],
															  True, ""
															  ], "middle-initial" -> Which[
																						   StringTake[ae1, -2] === "MN" && Length[a] == 1, 
																						   First[a],
																						   StringTake[ae2, -2] === "MN" && Length[b] == 1, 
																						   First[b],
																						   StringTake[ae3, -2] === "MN" && Length[c] == 1, 
																						   First[c],
																						   True, ""
																						   ], "last-name" -> Which[
																												   StringTake[ae1, -2] === "LN" && Length[a] == 1, 
																												   First[a],
																												   StringTake[ae2, -2] === "LN" && Length[b] == 1, 
																												   First[b],
																												   StringTake[ae3, -2] === "LN" && Length[c] == 1, 
																												   First[c],
																												   True, ""
																												   ]}, {Which[
																															  StringTake[ae1, -2] === "LN" && Length[a] == 1, 
																															  First[a],
																															  StringTake[ae2, -2] === "LN" && Length[b] == 1, 
																															  First[b],
																															  StringTake[ae3, -2] === "LN" && Length[c] == 1, 
																															  First[c],
																															  True, ""
																															  ] <> ", " <> Which[
																																				 StringTake[ae1, -2] === "FN" && Length[a] == 1, 
																																				 First[a],
																																				 StringTake[ae2, -2] === "FN" && Length[b] == 1, 
																																				 First[b],
																																				 StringTake[ae3, -2] === "FN" && Length[c] == 1, 
																																				 First[c],
																																				 True, ""
																																				 ] <> " " <> Which[
																																								   StringTake[ae1, -2] === "MN" && Length[a] == 1, 
																																								   First[a],
																																								   StringTake[ae2, -2] === "MN" && Length[b] == 1, 
																																								   First[b],
																																								   StringTake[ae3, -2] === "MN" && Length[c] == 1, 
																																								   First[c],
																																								   True, ""
																																								   ]}]
			  }]
		  }],
		  post},
	  currXpr
	  ]
   ) //.
  currXpr : {pre___,
	  XMLElement[
				 ae1_String?(StringMatchQ[#, "Author*N"] || 
							 StringMatchQ[#, "Editor*N"] &), {}, a_List], 
	  XMLElement[
				 ae2_String?(StringMatchQ[#, "Author*N"] || 
							 StringMatchQ[#, "Editor*N"] &), {}, b_List], post___} :> 
  If[
	 StringDrop[ae1, -2] === StringDrop[ae2, -2],
	 {pre,
		 XMLElement["contributors", {}, {
			 XMLElement[Which[
							  StringTake[ae1, 6] === "Editor", "secondary-authors",
							  True, "authors"
							  ], {}, {
				 XMLElement["author", {"first-name" -> Which[
															 StringTake[ae1, -2] === "FN" && Length[a] == 1, 
															 First[a],
															 StringTake[ae2, -2] === "FN" && Length[b] == 1, 
															 First[b],
															 True, ""
															 ], "middle-initial" -> Which[
																						  StringTake[ae1, -2] === "MN" && Length[a] == 1, 
																						  First[a],
																						  StringTake[ae2, -2] === "MN" && Length[b] == 1, 
																						  First[b],
																						  True, ""
																						  ], "last-name" -> Which[
																												  StringTake[ae1, -2] === "LN" && Length[a] == 1, 
																												  First[a],
																												  StringTake[ae2, -2] === "LN" && Length[b] == 1, 
																												  First[b],
																												  True, ""
																												  ]}, {Which[
																															 StringTake[ae1, -2] === "LN" && Length[a] == 1, 
																															 First[a],
																															 StringTake[ae2, -2] === "LN" && Length[b] == 1, 
																															 First[b],
																															 True, ""
																															 ] <> ", " <> Which[
																																				StringTake[ae1, -2] === "FN" && Length[a] == 1, 
																																				First[a],
																																				StringTake[ae2, -2] === "FN" && Length[b] == 1, 
																																				First[b],
																																				True, ""
																																				] <> " " <> Which[
																																								  StringTake[ae1, -2] === "MN" && Length[a] == 1, 
																																								  First[a],
																																								  StringTake[ae2, -2] === "MN" && Length[b] == 1, 
																																								  First[b],
																																								  True, ""
																																								  ]}]
			 }]
         }],
		 post},
	 currXpr
	 ]
  ) //.
 currXpr : {pre___,
	XMLElement[
			   ae1_String?(StringMatchQ[#, "Author*N"] || 
						   StringMatchQ[#, "Editor*N"] || # === "EditorGroup" &), {}, 
			   a_List], post___} :>
 {pre,
    XMLElement["contributors", {}, {
		XMLElement[Which[
						 StringTake[ae1, 6] === "Editor" || ae1 === "EditorGroup", 
						 "secondary-authors",
						 True, "authors"
						 ], {}, {
			XMLElement["author", {"first-name" -> Which[
														StringTake[ae1, -2] === "FN" && Length[a] == 1, First[a],
														True, ""
														], "middle-initial" -> Which[
																					 StringTake[ae1, -2] === "MN" && Length[a] == 1, First[a],
																					 True, ""
																					 ], "last-name" -> Which[
																											 StringTake[ae1, -2] === "LN" && Length[a] == 1, First[a],
																											 True, ""
																											 ]}, {Which[
																														
																														StringTake[ae1, -2] === "LN" || 
																														ae1 === "EditorGroup" && Length[a] == 1, First[a],
																														True, ""
																														] <> ", " <> Which[
																																		   StringTake[ae1, -2] === "FN" && Length[a] == 1, First[a],
																																		   True, ""
																																		   ] <> " " <> Which[
																																							 StringTake[ae1, -2] === "MN" && Length[a] == 1, First[a],
																																							 True, ""
																																							 ]}]
        }]
	}],
    post}
 
 ConvertTitle[xmlXpr_] := 
 xmlXpr //. 
 XMLElement[
			ttType_String?(StringMatchQ[#, "BookTitle"] || 
						   StringMatchQ[#, "JournalTitle"] || 
						   StringMatchQ[#, "Title"] &), {}, titleTxt_List] :>
 Which[
	   ttType === "BookTitle" || ttType === "Title",
	   XMLElement["titles", {}, {XMLElement["title", {}, titleTxt]}],
	   ttType === "JournalTitle",
	   XMLElement[
				  "titles", {}, {XMLElement["secondary-title", {}, titleTxt]}]
	   ]
 
 ConvertDate[xmlXpr_] :=
 (
  (xmlXpr /. {pre___, 
	XMLElement[calFld1 : ("Day" | "Month" | "Year"), {}, a_List], 
	XMLElement[calFld2 : ("Day" | "Month" | "Year"), {}, b_List], 
	XMLElement[calFld3 : ("Day" | "Month" | "Year"), {}, c_List],
	post___} :> {pre,
        XMLElement["dates", {}, {
			XMLElement["year", {"year" -> Which[
												calFld1 === "Year" && Length[a] == 1, First[a],
												calFld2 === "Year" && Length[b] == 1, First[b],
												calFld3 === "Year" && Length[c] == 1, First[c],
												True, ""
												]}, {Which[
														   calFld1 === "Year" && Length[a] == 1, First[a],
														   calFld2 === "Year" && Length[b] == 1, First[b],
														   calFld3 === "Year" && Length[c] == 1, First[c],
														   True, ""
														   ]}],
			
			XMLElement[
					   "pub-dates", {}, {XMLElement["date", {"month" -> Which[
																			  calFld1 === "Month" && Length[a] == 1, First[a],
																			  calFld2 === "Month" && Length[b] == 1, First[b],
																			  calFld3 === "Month" && Length[c] == 1, First[c],
																			  True, ""
																			  ],
				"day" -> Which[
							   calFld1 === "Day" && Length[a] == 1, First[a],
							   calFld2 === "Day" && Length[b] == 1, First[b],
							   calFld3 === "Day" && Length[c] == 1, First[c],
							   True, ""
							   ]}, {
					Which[
						  calFld1 === "Month" && Length[a] == 1, First[a],
						  calFld2 === "Month" && Length[b] == 1, First[b],
						  calFld3 === "Month" && Length[c] == 1, First[c],
						  True, ""
						  ] <> " " <> Which[
											calFld1 === "Day" && Length[a] == 1, First[a],
											calFld2 === "Day" && Length[b] == 1, First[b],
											calFld3 === "Day" && Length[c] == 1, First[c],
											True, ""
											]
				}]}]
		}], post}
   ) /.
  {pre___, 
      XMLElement[calFld1 : ("Day" | "Month" | "Year"), {}, a_List], 
      XMLElement[calFld2 : ("Day" | "Month" | "Year"), {}, b_List],
      post___} :> {pre,
		  XMLElement["dates", {}, {
			  XMLElement["year", {"year" -> Which[
												  calFld1 === "Year" && Length[a] == 1, First[a],
												  calFld2 === "Year" && Length[b] == 1, First[b],
												  True, ""
												  ]}, {Which[
															 calFld1 === "Year" && Length[a] == 1, First[a],
															 calFld2 === "Year" && Length[b] == 1, First[b],
															 True, ""
															 ]}],
			  XMLElement[
						 "pub-dates", {}, {XMLElement["date", {"month" -> Which[
																				calFld1 === "Month" && Length[a] == 1, First[a],
																				calFld2 === "Month" && Length[b] == 1, First[b],
																				True, ""
																				],
				  "day" -> Which[
								 calFld1 === "Day" && Length[a] == 1, First[a],
								 calFld2 === "Day" && Length[b] == 1, First[b],
								 True, ""
								 ]}, {
					  Which[
							calFld1 === "Month" && Length[a] == 1, First[a],
							calFld2 === "Month" && Length[b] == 1, First[b],
							True, ""
							] <> " " <> Which[
											  calFld1 === "Day" && Length[a] == 1, First[a],
											  calFld2 === "Day" && Length[b] == 1, First[b],
											  True, ""
											  ]
				  }]}]
		  }], post}
  ) /.
 {pre___, 
    XMLElement[calFld1 : ("Day" | "Month" | "Year"), {}, a_List], 
    post___} :> {pre,
		XMLElement["dates", {}, {
			XMLElement["year", {"year" -> Which[
												calFld1 === "Year" && Length[a] == 1, First[a],
												True, ""
												]}, {Which[
														   calFld1 === "Year" && Length[a] == 1, First[a],
														   True, ""
														   ]}],
			XMLElement[
					   "pub-dates", {}, {XMLElement["date", {"month" -> Which[
																			  calFld1 === "Month" && Length[a] == 1, First[a],
																			  True, ""
																			  ],
				"day" -> Which[
							   calFld1 === "Day" && Length[a] == 1, First[a],
							   True, ""
							   ]}, {
					Which[
						  calFld1 === "Month" && Length[a] == 1, First[a],
						  True, ""
						  ] <> " " <> Which[
											calFld1 === "Day" && Length[a] == 1, First[a],
											True, ""
											]
				}]}]
		}], post}
 
 ConvertPublisherInfo[xmlXpr_] :=
 (
  (xmlXpr /. {pre___, 
	XMLElement[pubFld1 : ("City" | "State" | "Country"), {}, 
			   a_List], 
	XMLElement[pubFld2 : ("City" | "State" | "Country"), {}, 
			   b_List], 
	XMLElement[pubFld3 : ("City" | "State" | "Country"), {}, 
			   c_List],
	post___} :> {pre,
        XMLElement["pub-location", {}, {Which[
											  pubFld1 === "City" && Length[a] == 1, First[a] <> ", ",
											  pubFld2 === "City" && Length[b] == 1, First[b] <> ", ",
											  pubFld3 === "City" && Length[c] == 1, First[c] <> ", ",
											  True, ""
											  ] <> Which[
														 pubFld1 === "State" && Length[a] == 1, First[a] <> ", ",
														 pubFld2 === "State" && Length[b] == 1, First[b] <> ", ",
														 pubFld3 === "State" && Length[c] == 1, First[c] <> ", ",
														 True, ""
														 ] <> Which[
																	pubFld1 === "Country" && Length[a] == 1, First[a],
																	pubFld2 === "Country" && Length[b] == 1, First[b],
																	pubFld3 === "Country" && Length[c] == 1, First[c],
																	True, ""
																	]
		}],
        post}
   ) /.
  {pre___, 
      XMLElement[pubFld1 : ("City" | "State" | "Country"), {}, 
				 a_List], 
      XMLElement[pubFld2 : ("City" | "State" | "Country"), {}, b_List],
      post___} :> {pre,
		  XMLElement["pub-location", {}, {Which[
												pubFld1 === "City" && Length[a] == 1, First[a] <> ", ",
												pubFld2 === "City" && Length[b] == 1, First[b] <> ", ",
												True, ""
												] <> Which[
														   pubFld1 === "State" && Length[a] == 1, First[a] <> ", ",
														   pubFld2 === "State" && Length[b] == 1, First[b] <> ", ",
														   True, ""
														   ] <> Which[
																	  pubFld1 === "Country" && Length[a] == 1, First[a],
																	  pubFld2 === "Country" && Length[b] == 1, First[b],
																	  True, ""
																	  ]
		  }],
		  post}
  ) /.
 {pre___, 
    XMLElement[pubFld1 : ("City" | "State" | "Country"), {}, a_List], 
    post___} :> {pre,
		XMLElement["pub-location", {}, {Which[
											  pubFld1 === "City" && Length[a] == 1, First[a],
											  True, ""
											  ] <> Which[
														 pubFld1 === "State" && Length[a] == 1, First[a],
														 True, ""
														 ] <> Which[
																	pubFld1 === "Country" && Length[a] == 1, First[a],
																	True, ""
																	]
		}],
		post}
 
 Options[ConvertFlatFields] = {
	"Database" -> "remote-database-name",
	"JournalVolume" -> "volume",
	"JournalNumber" -> "number",
	"Number" -> "number",
	"Notes" -> "notes",
	"NoteText" -> "notes",
	"Publisher" -> "publisher",
	"RetrievalDate" -> "access-date",
	"ReviewNumber" -> "accession-num",
	"ReviewVolume" -> "orig-pub",
	"School" -> "publisher",
	"SourceOfAbstract" -> "remote-source",
	"PublicationStatus" -> "reviewed-item",
	"ISBN" -> "isbn",
	"ISSN" -> "isbn"
};
 
 ConvertFlatFields[xmlXpr_, addedConvRules___?OptionQ] :=
 Module[
		{
			convRules = Options[ConvertFlatFields], convRulesNew
		},
		
		convRulesNew = Join[convRules, {addedConvRules}];
		
		xmlXpr //.
		
		XMLElement["record", recOps_, recCont_] :> 
		XMLElement["record", recOps, 
				   recCont //.
				   
				   XMLElement[styNm_String, styOps_, cont_] :> 
				   XMLElement[styNm /. convRulesNew, styOps, cont]]
		]
 
 
 ConvertPageRange[
				  xmlXpr_] :=
 (xmlXpr /. {pre___, 
	XMLElement[pg1 : ("FirstPage" | "LastPage"), {}, a_List], 
	XMLElement[pg2 : ("FirstPage" | "LastPage"), {}, b_List],
	post___} :> {pre,
		XMLElement["pages", {"start" -> Which[
											  pg1 === "FirstPage" && Length[a] == 1, First[a],
											  pg2 === "FirstPage" && Length[b] == 1, First[b],
											  True, ""
											  ],
			"end" -> Which[
						   pg1 === "LastPage" && Length[a] == 1, First[a],
						   pg2 === "LastPage" && Length[b] == 1, First[b],
						   True, ""
						   ]}, {Which[
									  pg1 === "FirstPage" && Length[a] == 1, First[a],
									  pg2 === "FirstPage" && Length[b] == 1, First[b],
									  True, ""
									  ] <> "-" <> Which[
														pg1 === "LastPage" && Length[a] == 1, First[a],
														pg2 === "LastPage" && Length[b] == 1, First[b],
														True, ""
														]
			}],
		post}
  ) /.
 {pre___, 
    XMLElement[pg1 : ("FirstPage" | "LastPage"), {}, a_List],
    post___} :> {pre,
		XMLElement["pages", {"start" -> Which[
											  pg1 === "FirstPage" && Length[a] == 1, First[a],
											  True, ""
											  ],
			"end" -> Which[
						   pg1 === "LastPage" && Length[a] == 1, First[a],
						   True, ""
						   ]}, {Which[
									  Length[a] == 1, First[a],
									  True, ""
									  ]
			}],
		post}
 
 ConvertURL[xmlXpr_] :=
 
 xmlXpr //. {pre___, XMLElement[("URL" | "UMI#"), styOps_, cont_], 
 post___} :> {pre, XMLElement["urls", {}, {
	XMLElement["related-urls", {}, {
        XMLElement["url", styOps, cont]
	}]
}], post}
 
 FixStyledContent[xpr_] :=
 
 xpr //. XMLElement[sty_, opts_, cont : {pre___,
 XMLElement["style", _, {_String}],
 post___
 }?(Length[Cases[#, _String, {0, 1}]] > 0 &)] :> 
 XMLElement[sty, opts, If[Head[#] === String,
						  XMLElement["style", {}, {#}],
#] & /@ cont]
						  
						  FixSecondaryTitle[xpr_] :=
						  
						  xpr /. XMLElement["record", opts_List, {pre___,
	titl1 : XMLElement["titles", _, {
        XMLElement["title", _, _]
	}],
	mid___,
	XMLElement[a : "titles", b_, {
		XMLElement["title", c_, d_]
	}],
	post___}] :> XMLElement["record", opts,
							{pre, titl1, mid,
								XMLElement[a, b, {
									XMLElement["secondary-title", c, d]
								}], post}
							]
						  
						  RemoveMathML[xpr_] :=
						  
						  xpr //. XMLElement[
						  "math", {{_String, 
	"xmlns"} -> _String?(StringMatchQ[#, "http://*/MathML"] &)}, 
						  cont_List] :> XMLElement["style", {}, {GetXMLCont[cont]}]
			
			GetXMLCont[xpr_] :=
			xpr /. XMLElement[_String, _, cont_List] :> Table[
															  If[Head[cont[[i]]] === XMLElement, GetXMLCont[cont[[i]]], 
																 cont[[i]]],
															  {i, Length[cont]}] // Flatten // StringJoin
			
			
			(*"style" text is interwoven between unformatted strings. To be valid \
			 the unformatted strings must also be formatted*)
			
			FixPlainText[xpr_] :=
			
			xpr //. XMLElement[sty_, opts_, 
			cont : {___, XMLElement["style", _, {_String}], ___}] :> 
 XMLElement[sty, opts,
			Table[
				  Which[
						Head[cont[[i]]] === String && i == 1,
						XMLElement["style", {}, {cont[[i]] <> " "}],
						Head[cont[[i]]] === String && i == Length[cont],   
						XMLElement["style", {}, {" " <> cont[[i]]}],
						Head[cont[[i]]] === String,
						XMLElement["style", {}, {" " <> cont[[i]] <> " "}],
						True,
						cont[[i]]
						], {i, Length[cont]}]]
 
 
 (* Functions CheckRefTypes and CheckFieldTypes are used to verify whether refTypes and data fields are able to be converted correctly. If not, a warning message appears. But still needs to be implemented.*)
 
 CheckRefTypes[xmlXpr_] := Module[
								  {
									  refTypesLst,
									  unsupportedRefTypesLst,
									  unsupportedRefTypesLstNoNull,
									  expectedRefTypes = {"Bill", "Case", "Hearing", "Statute", 
										  "Electronic Article", "Article", "Magazine Article", 
										  "Newspaper Article", "Artwork", "Blog", "WholeBook", 
										  "Edited Book", "Electronic Book", "Film or Broadcast", "Catalog",
										  "Personal Communication", "Aggregated Database", "OnlineSource",
										  "Dataset", "Dictionary", "Government Document", "Encyclopedia", 
										  "Equation", "Figure", "Grant", "Manuscript", "Map", 
										  "Audiovisual Material", "Online Multimedia", "Web Page", 
										  "Pamphlet", "Conference Paper", "Patent", 
										  "Conference Proceedings", "Computer Program", 
										  "Legal Rule or Regulation", "Report", "InBook", 
										  "Electronic Book Section", "Serial", "Standard", 
										  "Chart or Table", "Ancient Text", "Thesis", "Classical Work", 
										  "Preprint", "Note"}
								  },
								  
								  refTypesLst = 
								  Cases[xmlXpr, {Rule["id", _String], 
	Rule["type", refType_String]} :> refType, \[Infinity]] // Union;
								  
								  unsupportedRefTypesLst = If[Not@MemberQ[expectedRefTypes, #],
#,
															  Null] & /@ refTypesLst;
								  
								  unsupportedRefTypesLstNoNull = 
								  Select[unsupportedRefTypesLst, # =!= Null &];
								  
								  If[Length[unsupportedRefTypesLstNoNull] > 0,
									 Message[CheckRefTypes::unsupportedRefTypes, 
											 unsupportedRefTypesLstNoNull]
									 ]
								  ]
 
 CheckRefTypes::unsupportedRefTypes = 
 "WARNING: The imported database contains the following reference types that \
 are not supported: `1`. They will be mapped to a generic EndNote \
 reference type.";
 
 CheckFieldTypes[xmlXpr_] := Module[
									{
										refTypesLst,
										unsupportedRefTypesLst,
										unsupportedRefTypesLstNoNull,
										flatflds = Cases[Options[ConvertFlatFields], Rule[a_String, _String] :> a, 1],
										titles = {"BookTitle", "JournalTitle", "Title"},
										dates = {"Day", "Month", "Year"},
										pubinfo = {"City", "State", "Country"},
										pgrange = {"FirstPage" , "LastPage"},
										url = {"URL", "UMI#"},
										editors = {"EditorGroup"},
										expectedRefTypes
									},
									
									Catch[
										  If[Head[flatflds] =!= List || Length[flatflds] == 0,
											 (
											  Message[CheckFieldTypes::aborted, "Options[ConvertFlatFields]", 
													  Options[ConvertFlatFields]];
											  Throw[$Failed]
											  )];
										  
										  expectedRefTypes = Join[flatflds, titles, dates, pubinfo, pgrange, url, editors];
										  
										  refTypesLst = Cases[xmlXpr, {Rule["type", refType_String]} :> 
												refType, \[Infinity]] // Union;
										  
										  unsupportedRefTypesLst = If[Not@MemberQ[expectedRefTypes, #] && Not@StringMatchQ[#, ("Author" | "Editor") ~~ 
														  DigitCharacter .. ~~ ("F" | "M" | "L") ~~ "N"],
											 #,
											 Null] & /@ refTypesLst;
										  
										  unsupportedRefTypesLstNoNull = Select[unsupportedRefTypesLst, # =!= Null &];
										  
										  If[Head[unsupportedRefTypesLstNoNull] =!= List,
											 (
											  Message[CheckFieldTypes::aborted, "unsupportedRefTypesLstNoNull",
													  unsupportedRefTypesLstNoNull];
											  Throw[$Failed]
											  )
											 ];
										  
										  If[Length[unsupportedRefTypesLstNoNull] > 0,
											 Message[CheckFieldTypes::unsupportedFieldTypes, 
													 unsupportedRefTypesLstNoNull]
											 ]
										  ]
									]
 
 CheckFieldTypes::unsupportedFieldTypes = 
 "The imported database contains the following data field types that \
 are not supported: `1`. The associated data will NOT be converted.";
 
 CheckFieldTypes::aborted = 
 "WARNING: Unable to complete CheckFieldTypes because `1` returned: \
 `2`. Data may be lost!!!"; 
 
 
 
End[]

EndPackage[]

